/*
 * Copyright 2019 Confluent Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.confluent.kafka.secretregistry.client;

import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;

import io.confluent.kafka.secretregistry.client.rest.RestService;
import io.confluent.kafka.secretregistry.client.rest.Versions;
import io.confluent.kafka.secretregistry.client.rest.entities.Secret;
import io.confluent.kafka.secretregistry.client.rest.exceptions.RestClientException;
import io.confluent.kafka.secretregistry.client.security.basicauth.BasicAuthCredentialProvider;
import io.confluent.kafka.secretregistry.client.security.basicauth.BasicAuthCredentialProviderFactory;

/**
 * Thread-safe Secret Registry Client with client side caching.
 */
public class CachedSecretRegistryClient implements SecretRegistryClient {

  private final RestService restService;
  private final int identityMapCapacity;
  // TODO use caffeine?
  private final Map<String, Map<String, Map<Integer, String>>> cache;

  static final Map<String, String> DEFAULT_REQUEST_PROPERTIES;

  static {
    DEFAULT_REQUEST_PROPERTIES = new HashMap<String, String>();
    DEFAULT_REQUEST_PROPERTIES.put("Content-Type", Versions.SECRET_REGISTRY_V1_JSON_WEIGHTED);
  }

  public CachedSecretRegistryClient(String baseUrl, int identityMapCapacity) {
    this(new RestService(baseUrl), identityMapCapacity);
  }

  public CachedSecretRegistryClient(List<String> baseUrls, int identityMapCapacity) {
    this(new RestService(baseUrls), identityMapCapacity);
  }

  public CachedSecretRegistryClient(RestService restService, int identityMapCapacity) {
    this(restService, identityMapCapacity, null);
  }

  public CachedSecretRegistryClient(
      String baseUrl,
      int identityMapCapacity,
      Map<String, ?> originals
  ) {
    this(baseUrl, identityMapCapacity, originals, null);
  }

  public CachedSecretRegistryClient(
      List<String> baseUrls,
      int identityMapCapacity,
      Map<String, ?> originals
  ) {
    this(baseUrls, identityMapCapacity, originals, null);
  }

  public CachedSecretRegistryClient(
      RestService restService,
      int identityMapCapacity,
      Map<String, ?> configs
  ) {
    this(restService, identityMapCapacity, configs, null);
  }

  public CachedSecretRegistryClient(
      String baseUrl,
      int identityMapCapacity,
      Map<String, ?> originals,
      Map<String, String> httpHeaders
  ) {
    this(new RestService(baseUrl), identityMapCapacity, originals, httpHeaders);
  }

  public CachedSecretRegistryClient(
      List<String> baseUrls,
      int identityMapCapacity,
      Map<String, ?> originals,
      Map<String, String> httpHeaders
  ) {
    this(new RestService(baseUrls), identityMapCapacity, originals, httpHeaders);
  }

  public CachedSecretRegistryClient(
      RestService restService,
      int identityMapCapacity,
      Map<String, ?> configs,
      Map<String, String> httpHeaders
  ) {
    this.identityMapCapacity = identityMapCapacity;
    this.cache = new HashMap<>();
    this.restService = restService;
    configureRestService(configs, httpHeaders);
  }

  private void configureRestService(Map<String, ?> configs, Map<String, String> httpHeaders) {
    if (httpHeaders != null) {
      restService.setHttpHeaders(httpHeaders);
    }

    if (configs != null) {

      String credentialSourceConfig =
          (String) configs.get(SecretRegistryClientConfig.BASIC_AUTH_CREDENTIALS_SOURCE);

      if (credentialSourceConfig != null && !credentialSourceConfig.isEmpty()) {
        BasicAuthCredentialProvider basicAuthCredentialProvider =
            BasicAuthCredentialProviderFactory.getBasicAuthCredentialProvider(
                credentialSourceConfig,
                configs
            );

        restService.setBasicAuthCredentialProvider(basicAuthCredentialProvider);
      }
    }
  }

  @Override
  public synchronized int register(String path, String key, String secret)
      throws IOException, RestClientException {
    return restService.registerSecret(secret, path, key);
  }

  @Override
  public Collection<String> listAllPaths() throws IOException, RestClientException {
    return restService.listAllPaths();
  }

  @Override
  public Collection<String> listAllKeys(String path) throws IOException, RestClientException {
    return restService.listAllKeys(path);
  }

  @Override
  public Collection<Integer> listAllVersions(String path, String key)
      throws IOException, RestClientException {
    return restService.listAllVersions(path, key);
  }

  @Override
  public List<SecretMetadata> getAllLatest(String path)
      throws IOException, RestClientException {
    return restService.getAllLatest(path).stream()
        .map(SecretMetadata::new)
        .collect(Collectors.toList());
  }

  @Override
  public List<SecretMetadata> getAllVersions(String path, String key)
      throws IOException, RestClientException {
    return restService.getAllVersions(path, key).stream()
        .map(SecretMetadata::new)
        .collect(Collectors.toList());
  }

  @Override
  public SecretMetadata getLatestSecretMetadata(String path, String key)
      throws IOException, RestClientException {
    Secret secret = restService.getLatestVersion(path, key);
    return secret != null ? new SecretMetadata(secret) : null;
  }

  @Override
  public SecretMetadata getSecretMetadata(String path, String key, int version)
      throws IOException, RestClientException {
    Secret secret = restService.getVersion(path, key, version);
    return secret != null ? new SecretMetadata(secret) : null;
  }

  @Override
  public void deletePath(String path) throws IOException, RestClientException {
    deletePath(DEFAULT_REQUEST_PROPERTIES, path);
  }

  @Override
  public void deletePath(
      Map<String, String> requestProperties, String path
  )
      throws IOException, RestClientException {
    Objects.requireNonNull(path, "path");
    restService.deletePath(requestProperties, path);
  }

  @Override
  public void deleteKey(String path, String key) throws IOException, RestClientException {
    deleteKey(DEFAULT_REQUEST_PROPERTIES, path, key);
  }

  @Override
  public void deleteKey(
      Map<String, String> requestProperties, String path, String key
  )
      throws IOException, RestClientException {
    Objects.requireNonNull(path, "path");
    restService.deleteKey(requestProperties, path, key);
  }

  @Override
  public void deleteVersion(String path, String key, String version)
      throws IOException, RestClientException {
    deleteVersion(DEFAULT_REQUEST_PROPERTIES, path, key, version);
  }

  @Override
  public void deleteVersion(
      Map<String, String> requestProperties,
      String path,
      String key,
      String version
  )
      throws IOException, RestClientException {
    restService.deleteVersion(requestProperties, path, key, version);
  }

  @Override
  public void reset() {
    cache.clear();
  }
}
