/*
 * Copyright 2019 Confluent Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.confluent.kafka.secretregistry.client;

import static io.confluent.kafka.secretregistry.client.CachedSecretRegistryClient.DEFAULT_REQUEST_PROPERTIES;

import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import io.confluent.kafka.secretregistry.client.rest.exceptions.RestClientException;

/**
 * Mock implementation of SecretRegistryClient that can be used for tests. This version is NOT
 * thread safe. Secret data is stored in memory and is not persistent or shared across instances.
 */
public class MockSecretRegistryClient implements SecretRegistryClient {

  private final Map<String, Map<String, Map<Integer, String>>> cache;

  public MockSecretRegistryClient() {
    cache = new ConcurrentHashMap<>();
  }

  @Override
  public synchronized int register(String path, String key, String secret)
      throws IOException, RestClientException {
    Map<String, Map<Integer, String>> keys = cache.computeIfAbsent(
        path, k -> new ConcurrentHashMap<>());
    Map<Integer, String> versions = keys.computeIfAbsent(key, k -> new ConcurrentHashMap<>());
    int version = versions.size() + 1;
    versions.put(version, secret);
    return version;
  }

  @Override
  public Collection<String> listAllPaths() throws IOException, RestClientException {
    return cache.keySet();
  }

  @Override
  public Collection<String> listAllKeys(String path) throws IOException, RestClientException {
    return cache.getOrDefault(path, Collections.emptyMap()).keySet();
  }

  @Override
  public Collection<Integer> listAllVersions(String path, String key)
      throws IOException, RestClientException {
    return cache.getOrDefault(path, Collections.emptyMap())
        .getOrDefault(key, Collections.emptyMap())
        .keySet();
  }

  @Override
  public List<SecretMetadata> getAllLatest(String path)
      throws IOException, RestClientException {
    return cache.getOrDefault(path, Collections.emptyMap())
        .entrySet().stream()
        .map(e -> {
          String key = e.getKey();
          Map<Integer, String> versions = e.getValue();
          return versions.entrySet().stream()
              .max((e1, e2) -> e1.getKey().compareTo(e2.getKey()))
              .map(o -> new SecretMetadata(path, key, o.getKey(), o.getValue()));
        })
        .flatMap(o -> o.map(Stream::of).orElseGet(Stream::empty))
        .collect(Collectors.toList());
  }

  @Override
  public List<SecretMetadata> getAllVersions(String path, String key)
      throws IOException, RestClientException {
    return cache.getOrDefault(path, Collections.emptyMap())
        .getOrDefault(key, Collections.emptyMap())
        .entrySet().stream()
        .map(e -> new SecretMetadata(path, key, e.getKey(), e.getValue()))
        .collect(Collectors.toList());
  }

  @Override
  public SecretMetadata getLatestSecretMetadata(String path, String key)
      throws IOException, RestClientException {
    return cache.getOrDefault(path, Collections.emptyMap())
        .getOrDefault(key, Collections.emptyMap())
        .entrySet().stream()
        .max((e1, e2) -> e1.getKey().compareTo(e2.getKey()))
        .map(o -> new SecretMetadata(path, key, o.getKey(), o.getValue()))
        .orElse(null);
  }

  @Override
  public SecretMetadata getSecretMetadata(String path, String key, int version)
      throws IOException, RestClientException {
    String secret = cache.getOrDefault(path, Collections.emptyMap())
        .getOrDefault(key, Collections.emptyMap())
        .getOrDefault(version, null);
    return secret != null ? new SecretMetadata(path, key, version, secret) : null;
  }

  @Override
  public void deletePath(String path) throws IOException, RestClientException {
    deletePath(DEFAULT_REQUEST_PROPERTIES, path);
  }

  @Override
  public void deletePath(
      Map<String, String> requestProperties, String path
  )
      throws IOException, RestClientException {
    cache.remove(path);
  }

  @Override
  public void deleteKey(String path, String key) throws IOException, RestClientException {
    deleteKey(DEFAULT_REQUEST_PROPERTIES, path, key);
  }

  @Override
  public void deleteKey(
      Map<String, String> requestProperties, String path, String key
  )
      throws IOException, RestClientException {
    cache.getOrDefault(path, new HashMap<>()).remove(key);
  }

  @Override
  public void deleteVersion(String path, String key, String version)
      throws IOException, RestClientException {
    deleteVersion(DEFAULT_REQUEST_PROPERTIES, path, key, version);
  }

  @Override
  public void deleteVersion(
      Map<String, String> requestProperties,
      String path,
      String key,
      String version
  )
      throws IOException, RestClientException {
    cache.getOrDefault(path, new HashMap<>())
        .getOrDefault(key, new HashMap<>())
        .remove(version);
  }

  @Override
  public void reset() {
    cache.clear();
  }
}
