/*
 * Copyright 2019 Confluent Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.confluent.kafka.secretregistry.client.config.provider;

import java.io.IOException;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.kafka.common.config.ConfigData;
import org.apache.kafka.common.config.provider.ConfigProvider;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import io.confluent.kafka.secretregistry.client.CachedSecretRegistryClient;
import io.confluent.kafka.secretregistry.client.SecretMetadata;
import io.confluent.kafka.secretregistry.client.SecretRegistryClient;
import io.confluent.kafka.secretregistry.client.rest.exceptions.RestClientException;

/**
 * An implementation of {@link ConfigProvider} that obtains secrets from the secret registry.
 */
public class SecretConfigProvider implements ConfigProvider {

  private static final Logger log = LoggerFactory.getLogger(SecretConfigProvider.class);

  private SecretRegistryClient secretRegistry;

  public void configure(Map<String, ?> configs) {
    SecretConfigProviderConfig providerConfig = new SecretConfigProviderConfig(configs);

    if (secretRegistry == null) {
      secretRegistry = new CachedSecretRegistryClient(
          providerConfig.getSecretRegistryUrls(),
          Integer.MAX_VALUE,
          configs,
          providerConfig.requestHeaders()
      );
    }
  }

  /**
   * Retrieves the data from the secret registry at the given path.
   *
   * @param path the path where the data resides
   * @return the configuration data
   */
  public ConfigData get(String path) {
    try {
      Map<String, String> data = secretRegistry.getAllLatest(path).stream()
          .collect(Collectors.toMap(SecretMetadata::getKey, SecretMetadata::getSecret));
      return new ConfigData(data);
    } catch (IOException | RestClientException e) {
      log.error("Could not obtain config data", e);
      throw new RuntimeException(e);
    }
  }

  /**
   * Retrieves the data with the given keys at the given path.
   *
   * @param path the path where the data resides
   * @param keys the keys whose values will be retrieved
   * @return the configuration data
   */
  public ConfigData get(String path, Set<String> keys) {
    try {
      Map<String, String> data = secretRegistry.getAllLatest(path).stream()
          .filter(s -> keys.contains(s.getKey()))
          .collect(Collectors.toMap(SecretMetadata::getKey, SecretMetadata::getSecret));
      return new ConfigData(data);
    } catch (IOException | RestClientException e) {
      log.error("Could not obtain config data", e);
      throw new RuntimeException(e);
    }
  }

  public void close() {
    secretRegistry.reset();
  }
}
