/*
 * Copyright 2019 Confluent Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.confluent.kafka.secretregistry.client.config.provider;

import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;

import org.apache.kafka.common.config.AbstractConfig;
import org.apache.kafka.common.config.ConfigDef;
import org.apache.kafka.common.config.ConfigDef.Importance;
import org.apache.kafka.common.config.ConfigDef.Type;

import io.confluent.kafka.secretregistry.client.SecretRegistryClientConfig;

/**
 * Base class for configs for serializers and deserializers, defining a few common configs and
 * defaults.
 */
public class SecretConfigProviderConfig extends AbstractConfig {

  /**
   * Configurations beginning with this prefix can be used to specify headers to include in requests
   * made to Secret Registry. For example, to include an {@code Authorization} header with a value
   * of {@code Bearer NjksNDIw}, use the following configuration:
   *
   * <p>{@code request.header.Authorization=Bearer NjksNDIw}
   */
  public static final String REQUEST_HEADER_PREFIX = "request.header.";

  public static final String SECRET_REGISTRY_URL_CONFIG = "secret.registry.url";
  public static final String
      SECRET_REGISTRY_URL_DOC =
      "Comma-separated list of URLs for secret registry instances that can be used to register "
      + "or look up secrets.";

  public static final String BASIC_AUTH_CREDENTIALS_SOURCE = SecretRegistryClientConfig
      .BASIC_AUTH_CREDENTIALS_SOURCE;
  public static final String BASIC_AUTH_CREDENTIALS_SOURCE_DEFAULT = "URL";
  public static final String BASIC_AUTH_CREDENTIALS_SOURCE_DOC =
      "Specify how to pick the credentials for Basic Auth header. "
      + "The supported values are URL, USER_INFO and SASL_INHERIT";

  public static final String USER_INFO_CONFIG =
      SecretRegistryClientConfig.USER_INFO_CONFIG;
  public static final String USER_INFO_DEFAULT = "";
  public static final String USER_INFO_DOC =
      "Specify the user info for Basic Auth in the form of {username}:{password}";

  public static ConfigDef baseConfigDef() {
    return new ConfigDef()
        .define(SECRET_REGISTRY_URL_CONFIG, Type.LIST,
                Importance.HIGH, SECRET_REGISTRY_URL_DOC)
        .define(BASIC_AUTH_CREDENTIALS_SOURCE, Type.STRING, BASIC_AUTH_CREDENTIALS_SOURCE_DEFAULT,
            Importance.MEDIUM, BASIC_AUTH_CREDENTIALS_SOURCE_DOC)
        .define(USER_INFO_CONFIG, Type.PASSWORD, USER_INFO_DEFAULT,
                Importance.MEDIUM, USER_INFO_DOC);
  }

  public SecretConfigProviderConfig(Map<?, ?> props) {
    super(baseConfigDef(), props);
  }

  public List<String> getSecretRegistryUrls() {
    return this.getList(SECRET_REGISTRY_URL_CONFIG);
  }

  public Map<String, String> requestHeaders() {
    return originalsWithPrefix(REQUEST_HEADER_PREFIX).entrySet().stream()
        .collect(Collectors.toMap(Map.Entry::getKey, entry -> Objects.toString(entry.getValue())));
  }

  public String basicAuthUserInfo() {
    return getString(USER_INFO_CONFIG);
  }
}
