/*
 * Copyright 2019 Confluent Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.confluent.kafka.secretregistry.client.rest;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.Base64;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLSocketFactory;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import io.confluent.kafka.secretregistry.client.rest.entities.ErrorMessage;
import io.confluent.kafka.secretregistry.client.rest.entities.Secret;
import io.confluent.kafka.secretregistry.client.rest.entities.requests.RegisterSecretRequest;
import io.confluent.kafka.secretregistry.client.rest.entities.requests.RegisterSecretResponse;
import io.confluent.kafka.secretregistry.client.rest.exceptions.RestClientException;
import io.confluent.kafka.secretregistry.client.rest.utils.UrlList;
import io.confluent.kafka.secretregistry.client.security.basicauth.BasicAuthCredentialProvider;

/**
 * Rest access layer for sending requests to the secret registry.
 */
public class RestService {

  private static final Logger log = LoggerFactory.getLogger(RestService.class);
  private static final TypeReference<RegisterSecretResponse> REGISTER_RESPONSE_TYPE =
      new TypeReference<RegisterSecretResponse>() {
      };
  private static final TypeReference<JsonNode> GET_SECRET_ONLY_BY_VERSION_RESPONSE_TYPE =
      new TypeReference<JsonNode>() {
      };
  private static final TypeReference<Secret> GET_SECRET_BY_VERSION_RESPONSE_TYPE =
      new TypeReference<Secret>() {
      };
  private static final TypeReference<List<Secret>> ALL_SECRETS_RESPONSE_TYPE =
      new TypeReference<List<Secret>>() {
      };
  private static final TypeReference<List<Integer>> ALL_VERSIONS_RESPONSE_TYPE =
      new TypeReference<List<Integer>>() {
      };
  private static final TypeReference<List<String>> ALL_NAMES_RESPONSE_TYPE =
      new TypeReference<List<String>>() {
      };
  private static final TypeReference<Integer> DELETE_VERSION_RESPONSE_TYPE =
      new TypeReference<Integer>() {
      };
  private static final TypeReference<? extends List<Integer>> DELETE_KEY_RESPONSE_TYPE =
      new TypeReference<List<Integer>>() {
      };
  private static final TypeReference<? extends Set<String>> DELETE_PATH_RESPONSE_TYPE =
      new TypeReference<Set<String>>() {
      };

  private static final int HTTP_CONNECT_TIMEOUT_MS = 60000;
  private static final int HTTP_READ_TIMEOUT_MS = 60000;

  private static final int JSON_PARSE_ERROR_CODE = 50005;
  private static ObjectMapper jsonDeserializer = new ObjectMapper();

  static final Map<String, String> DEFAULT_REQUEST_PROPERTIES;

  static {
    DEFAULT_REQUEST_PROPERTIES = new HashMap<String, String>();
    DEFAULT_REQUEST_PROPERTIES.put("Content-Type", Versions.SECRET_REGISTRY_V1_JSON_WEIGHTED);
  }

  private UrlList baseUrls;
  private SSLSocketFactory sslSocketFactory;
  private BasicAuthCredentialProvider basicAuthCredentialProvider;
  private Map<String, String> httpHeaders;

  public RestService(UrlList baseUrls) {
    this.baseUrls = baseUrls;
  }

  public RestService(List<String> baseUrls) {
    this(new UrlList(baseUrls));
  }

  public RestService(String baseUrlConfig) {
    this(parseBaseUrl(baseUrlConfig));
  }

  public void setSslSocketFactory(SSLSocketFactory sslSocketFactory) {
    this.sslSocketFactory = sslSocketFactory;
  }

  /**
   * @param requestUrl        HTTP connection will be established with this url.
   * @param method            HTTP method ("GET", "POST", "PUT", etc.)
   * @param requestBodyData   Bytes to be sent in the request body.
   * @param requestProperties HTTP header properties.
   * @param responseFormat    Expected format of the response to the HTTP request.
   * @param <T>               The type of the deserialized response to the HTTP request.
   * @return The deserialized response to the HTTP request, or null if no data is expected.
   */
  private <T> T sendHttpRequest(
      String requestUrl, String method, byte[] requestBodyData,
      Map<String, String> requestProperties,
      TypeReference<T> responseFormat
  )
      throws IOException, RestClientException {
    String requestData = requestBodyData == null
        ? "null"
        : new String(requestBodyData, StandardCharsets.UTF_8);
    log.debug(String.format("Sending %s with input %s to %s",
        method, requestData,
        requestUrl
    ));

    HttpURLConnection connection = null;
    try {
      URL url = new URL(requestUrl);
      connection = (HttpURLConnection) url.openConnection();

      connection.setConnectTimeout(HTTP_CONNECT_TIMEOUT_MS);
      connection.setReadTimeout(HTTP_READ_TIMEOUT_MS);

      setupSsl(connection);
      connection.setRequestMethod(method);
      setBasicAuthRequestHeader(connection);
      setCustomHeaders(connection);
      // connection.getResponseCode() implicitly calls getInputStream, so always set to true.
      // On the other hand, leaving this out breaks nothing.
      connection.setDoInput(true);

      for (Map.Entry<String, String> entry : requestProperties.entrySet()) {
        connection.setRequestProperty(entry.getKey(), entry.getValue());
      }

      connection.setUseCaches(false);

      if (requestBodyData != null) {
        connection.setDoOutput(true);
        OutputStream os = null;
        try {
          os = connection.getOutputStream();
          os.write(requestBodyData);
          os.flush();
        } catch (IOException e) {
          log.error("Failed to send HTTP request to endpoint: " + url, e);
          throw e;
        } finally {
          if (os != null) {
            os.close();
          }
        }
      }

      int responseCode = connection.getResponseCode();
      if (responseCode == HttpURLConnection.HTTP_OK) {
        InputStream is = connection.getInputStream();
        T result = jsonDeserializer.readValue(is, responseFormat);
        is.close();
        return result;
      } else if (responseCode == HttpURLConnection.HTTP_NO_CONTENT) {
        return null;
      } else {
        InputStream es = connection.getErrorStream();
        ErrorMessage errorMessage;
        try {
          errorMessage = jsonDeserializer.readValue(es, ErrorMessage.class);
        } catch (JsonProcessingException e) {
          errorMessage = new ErrorMessage(JSON_PARSE_ERROR_CODE, e.getMessage());
        }
        es.close();
        throw new RestClientException(errorMessage.getMessage(), responseCode,
            errorMessage.getErrorCode()
        );
      }

    } finally {
      if (connection != null) {
        connection.disconnect();
      }
    }
  }

  private void setupSsl(HttpURLConnection connection) {
    if (connection instanceof HttpsURLConnection && sslSocketFactory != null) {
      ((HttpsURLConnection) connection).setSSLSocketFactory(sslSocketFactory);
    }
  }

  private <T> T httpRequest(
      String path,
      String method,
      byte[] requestBodyData,
      Map<String, String> requestProperties,
      TypeReference<T> responseFormat
  )
      throws IOException, RestClientException {
    for (int i = 0, n = baseUrls.size(); i < n; i++) {
      String baseUrl = baseUrls.current();
      String requestUrl = buildRequestUrl(baseUrl, path);
      try {
        return sendHttpRequest(
            requestUrl,
            method,
            requestBodyData,
            requestProperties,
            responseFormat
        );
      } catch (IOException e) {
        baseUrls.fail(baseUrl);
        if (i == n - 1) {
          throw e; // Raise the exception since we have no more urls to try
        }
      }
    }
    throw new IOException("Internal HTTP retry error"); // Can't get here
  }

  // Visible for testing
  static String buildRequestUrl(String baseUrl, String path) {
    // Join base URL and path, collapsing any duplicate forward slash delimiters
    return baseUrl.replaceFirst("/$", "") + "/" + path.replaceFirst("^/", "");
  }

  public int registerSecret(String secretString, String path, String key)
      throws IOException, RestClientException {
    RegisterSecretRequest request = new RegisterSecretRequest();
    request.setSecret(secretString);
    return registerSecret(request, path, key);
  }

  public int registerSecret(RegisterSecretRequest registerSecretRequest, String path, String key)
      throws IOException, RestClientException {
    return registerSecret(DEFAULT_REQUEST_PROPERTIES, registerSecretRequest, path, key);
  }

  public int registerSecret(
      Map<String, String> requestProperties,
      RegisterSecretRequest registerSecretRequest, String path, String key
  )
      throws IOException, RestClientException {
    String url = String.format("/secret/paths/%s/keys/%s/versions", path, key);

    RegisterSecretResponse response = httpRequest(
        url, "POST",
        registerSecretRequest.toJson().getBytes(StandardCharsets.UTF_8),
        requestProperties,
        REGISTER_RESPONSE_TYPE
    );

    return response.getVersion();
  }

  public Secret getVersion(String path, String key, int version)
      throws IOException, RestClientException {
    return getVersion(DEFAULT_REQUEST_PROPERTIES, path, key, version);
  }

  public Secret getVersion(
      Map<String, String> requestProperties,
      String path, String key, int version
  )
      throws IOException, RestClientException {
    String url = String.format("/secret/paths/%s/keys/%s/versions/%d", path, key, version);

    Secret response = httpRequest(url, "GET", null, requestProperties,
        GET_SECRET_BY_VERSION_RESPONSE_TYPE
    );
    return response;
  }

  public Secret getLatestVersion(String path, String key)
      throws IOException, RestClientException {
    return getLatestVersion(DEFAULT_REQUEST_PROPERTIES, path, key);
  }

  public Secret getLatestVersion(
      Map<String, String> requestProperties,
      String path, String key
  )
      throws IOException, RestClientException {
    String url = String.format("/secret/paths/%s/keys/%s/versions/latest", path, key);

    Secret response = httpRequest(url, "GET", null, requestProperties,
        GET_SECRET_BY_VERSION_RESPONSE_TYPE
    );
    return response;
  }

  public String getVersionSecretOnly(String path, String key, int version)
      throws IOException, RestClientException {
    String url = String.format("/secret/paths/%s/keys/%s/versions/%d/secret", path, key, version);

    JsonNode response = httpRequest(url, "GET", null, DEFAULT_REQUEST_PROPERTIES,
        GET_SECRET_ONLY_BY_VERSION_RESPONSE_TYPE
    );
    return response.toString();
  }

  public String getLatestVersionSecretOnly(String path, String key)
      throws IOException, RestClientException {
    String url = String.format("/secret/paths/%s/keys/%s/versions/latest/secret", path, key);

    JsonNode response = httpRequest(url, "GET", null, DEFAULT_REQUEST_PROPERTIES,
        GET_SECRET_ONLY_BY_VERSION_RESPONSE_TYPE
    );
    return response.toString();
  }

  public List<Secret> getAllVersions(String path, String key)
      throws IOException, RestClientException {
    return getAllVersions(DEFAULT_REQUEST_PROPERTIES, path, key);
  }

  public List<Secret> getAllVersions(
      Map<String, String> requestProperties,
      String path, String key
  )
      throws IOException, RestClientException {
    String url = String.format("/secret/paths/%s/keys/%s", path, key);

    List<Secret> response = httpRequest(url, "GET", null, requestProperties,
        ALL_SECRETS_RESPONSE_TYPE
    );
    return response;
  }

  public List<Secret> getAllLatest(String path)
      throws IOException, RestClientException {
    return getAllLatest(DEFAULT_REQUEST_PROPERTIES, path);
  }

  public List<Secret> getAllLatest(
      Map<String, String> requestProperties,
      String path
  )
      throws IOException, RestClientException {
    String url = String.format("/secret/paths/%s", path);

    List<Secret> response = httpRequest(url, "GET", null, requestProperties,
        ALL_SECRETS_RESPONSE_TYPE
    );
    return response;
  }

  public List<Integer> listAllVersions(String path, String key)
      throws IOException, RestClientException {
    return listAllVersions(DEFAULT_REQUEST_PROPERTIES, path, key);
  }

  public List<Integer> listAllVersions(
      Map<String, String> requestProperties, String path, String key
  )
      throws IOException, RestClientException {
    String url = String.format("/secret/paths/%s/keys/%s/versions", path, key);
    List<Integer> response = httpRequest(url, "GET", null, requestProperties,
        ALL_VERSIONS_RESPONSE_TYPE
    );
    return response;
  }

  public List<String> listAllKeys(String path)
      throws IOException, RestClientException {
    return listAllKeys(DEFAULT_REQUEST_PROPERTIES, path);
  }

  public List<String> listAllKeys(Map<String, String> requestProperties, String path)
      throws IOException, RestClientException {
    String url = String.format("/secret/paths/%s", path);
    List<String> response = httpRequest(url, "GET", null, requestProperties,
        ALL_NAMES_RESPONSE_TYPE
    );
    return response;
  }

  public List<String> listAllPaths()
      throws IOException, RestClientException {
    return listAllPaths(DEFAULT_REQUEST_PROPERTIES);
  }

  public List<String> listAllPaths(Map<String, String> requestProperties)
      throws IOException, RestClientException {
    List<String> response = httpRequest("/secret/paths", "GET", null, requestProperties,
        ALL_NAMES_RESPONSE_TYPE
    );
    return response;
  }

  public int deleteVersion(
      Map<String, String> requestProperties,
      String path,
      String key,
      String version
  ) throws IOException, RestClientException {
    String url = String.format("/secret/paths/%s/keys/%s/versions/%s", path, key, version);

    int response = httpRequest(url, "DELETE", null, requestProperties,
        DELETE_VERSION_RESPONSE_TYPE
    );
    return response;
  }

  public List<Integer> deleteKey(
      Map<String, String> requestProperties,
      String path,
      String key
  ) throws IOException, RestClientException {
    String url = String.format("/secret/paths/%s/keys/%s", path, key);

    List<Integer> response = httpRequest(url, "DELETE", null, requestProperties,
        DELETE_KEY_RESPONSE_TYPE
    );
    return response;
  }

  public Set<String> deletePath(
      Map<String, String> requestProperties,
      String path
  ) throws IOException, RestClientException {
    String url = String.format("/secret/paths/%s", path);

    Set<String> response = httpRequest(url, "DELETE", null, requestProperties,
        DELETE_PATH_RESPONSE_TYPE
    );
    return response;
  }

  private static List<String> parseBaseUrl(String baseUrl) {
    List<String> baseUrls = Arrays.asList(baseUrl.split("\\s*,\\s*"));
    if (baseUrls.isEmpty()) {
      throw new IllegalArgumentException("Missing required secret registry url list");
    }
    return baseUrls;
  }

  public UrlList getBaseUrls() {
    return baseUrls;
  }

  private void setBasicAuthRequestHeader(HttpURLConnection connection) {
    String userInfo;
    if (basicAuthCredentialProvider != null
        && (userInfo = basicAuthCredentialProvider.getUserInfo(connection.getURL())) != null) {
      String authHeader = Base64.getEncoder().encodeToString(
          userInfo.getBytes(StandardCharsets.UTF_8));
      connection.setRequestProperty("Authorization", "Basic " + authHeader);
    }
  }

  private void setCustomHeaders(HttpURLConnection connection) {
    if (httpHeaders != null) {
      httpHeaders.forEach((k, v) -> connection.setRequestProperty(k, v));
    }
  }

  public void setBasicAuthCredentialProvider(
      BasicAuthCredentialProvider basicAuthCredentialProvider
  ) {
    this.basicAuthCredentialProvider = basicAuthCredentialProvider;
  }

  public void setHttpHeaders(Map<String, String> httpHeaders) {
    this.httpHeaders = httpHeaders;
  }

}
